#!/usr/bin/env python
# -*- coding: utf-8 -*-
import math
from enum import Enum
from py65emu.mmu import MMU
from py65emu.operation import Operation, OpCodes
from py65emu.debug import Disassembly


class FlagBit(Enum):
    """
    The status register stores 8 flags. Ive enumerated these here for ease
    of access. You can access the status register directly since its public.
    The bits have different interpretations depending upon the context and
    instruction being executed.
    """

    N = 1 << 7
    """
    Negative

    The negative flag is set if the result of the last operation had bit 7
    set to a one.
    """

    V = 1 << 6
    """
    Overflow

    The overflow flag is set during arithmetic operations if the result has
    yielded an invalid 2's complement result (e.g. adding to positive numbers
    and ending up with a negative result: 64 + 64 => -128). It is determined
    by looking at the carry between bits 6 and 7 and between bit 7 and the
    carry flag.
    """

    U = 1 << 5
    """
    Unused

    Is always set
    """

    B = 1 << 4
    """
    Break Command

    The break command bit is set when a BRK instruction has been executed and
    an interrupt has been generated to process it.
    """

    D = 1 << 3
    """
    Decimal Mode

    While the decimal mode flag is set the processor will obey the rules of
    Binary Coded Decimal (BCD) arithmetic during addition and subtraction. The
    flag can be explicitly set using 'Set Decimal Flag' (SED) and cleared with
    'Clear Decimal Flag' (CLD).
    """

    I = 1 << 2
    """
    IRQ Disable

    The interrupt disable flag is set if the program has executed a
    'Set Interrupt Disable' (SEI) instruction. While this flag is set the
    processor will not respond to interrupts from devices until it is cleared
    by a 'Clear Interrupt Disable' (CLI) instruction.
    """

    Z = 1 << 1
    """
    Zero

    The zero flag is set if the result of the last operation as was zero.
    """

    C = 1 << 0
    """
    Carry

    The carry flag is set if the last operation caused an overflow from bit 7
    of the result or an underflow from bit 0. This condition is set during
    arithmetic, comparison and during logical shifts. It can be explicitly set
    using the 'Set Carry Flag' (SEC) instruction and cleared with
    'Clear Carry Flag' (CLC).
    """


class Registers:
    """
    CPU registers

    The 6502 has only a small number of registers compared to other processor
    of the same era. This makes it especially challenging to program as
    algorithms must make efficient use of both registers and memory.
    """

    a: int
    """
    Accumulator - 8 bit

    The 8 bit accumulator is used all arithmetic and logical operations (with
    the exception of increments and decrements). The contents of the
    accumulator can be stored and retrieved either from memory or the stack.

    Most complex operations will need to use the accumulator for arithmetic
    and efficient optimisation of its use is a key feature of time
    critical routines.
    """

    x: int
    """
    General Purpose X - 8 bit

    The 8 bit index register is most commonly used to hold counters or offsets
    for accessing memory. The value of the X register can be loaded and saved
    in memory, compared with values held in memory or incremented and
    decremented.

    The X register has one special function. It can be used to get a copy of
    the stack pointer or change its value.
    """

    y: int
    """
    General Purpose Y

    The Y register is similar to the X register in that it is available for
    holding counter or offsets memory access and supports the same set of
    memory load, save and compare operations as wells as increments and
    decrements. It has no special functions.
    """

    s: int
    """
    Stack Pointer - 8 bit

    The processor supports a 256 byte stack located between $0100 and $01FF.
    The stack pointer is an 8 bit register and holds the low 8 bits of the next
    free location on the stack. The location of the stack is fixed and
    cannot be moved.

    Pushing bytes to the stack causes the stack pointer to be decremented.
    Conversely pulling bytes causes it to be incremented.

    The CPU does not detect if the stack is overflowed by excessive pushing or
    pulling operations and will most likely result in the program crashing.
    """

    pc: int
    """
    Program Counter - 16 bit

    The program counter is a 16 bit register which points to the next
    instruction to be executed. The value of program counter is modified
    automatically as instructions are executed.

    The value of the program counter can be modified by executing a jump, a
    relative branch or a subroutine call to another memory address or by
    returning from a subroutine or interrupt.
    """

    p: int
    """
    Flag Pointer - 8 bit - N|V|1|B|D|I|Z|C

    As instructions are executed a set of processor flags are set or clear to
    record the results of the operation. This flags and some additional control
    flags are held in a special status register. Each flag has a single bit
    within the register.

    Instructions exist to test the values of the various bits, to set or clear
    some of them and to push or pull the entire set to or from the stack.
    """

    def __init__(self, pc: int = 0x0000):
        """Init Registers

        :param int | None pc: Position of Program Counter
        """
        self.a = 0           # Accumulator
        self.x = 0           # General Purpose X
        self.y = 0           # General Purpose Y
        self.s = 0xFF        # Stack Pointer
        self.pc = pc         # Program Counter
        self.p = 0b00100100  # Flag Pointer
        #          NV1BDIZC

    def reset(self, pc: int = 0x0000):
        """Reset Registers

        :param int | None pc: Position of Program Counter
        """
        self.a = 0           # Accumulator
        self.x = 0           # General Purpose X
        self.y = 0           # General Purpose Y
        self.s = 0xFF        # Stack Pointer
        self.pc = pc         # Program Counter
        self.p = 0b00100100  # Flag Pointer - N|V|1|B|D|I|Z|C

    def getFlag(self, flag: FlagBit | int | str) -> bool:
        """Get flag value

        :param flag: One of `N`, `V`, `B`, `D`, `I`, `Z`, `C`
        :type flag: FlagBit or int or str
        :rtype: bool
        :return: Whether the flag is set or not
        """
        if isinstance(flag, FlagBit):
            pass
        elif isinstance(flag, str):
            flag = FlagBit[flag]
        elif isinstance(flag, int):
            flag = FlagBit(flag)

        return bool(self.p & flag.value)

    def setFlag(
        self,
        flag: FlagBit | int | str,
        v: bool | int = True
    ) -> None:
        """
        Set flag value

        :param flag: One of `N`, `V`, `B`, `D`, `I`, `Z`, `C`
        :type flag: FlagBit or int or str
        :param bool | int v: Flag is set or not
        """
        if isinstance(flag, FlagBit):
            pass
        elif isinstance(flag, str):
            flag = FlagBit[flag]
        elif isinstance(flag, int):
            flag = FlagBit(flag)

        if v > 0:
            self.p = self.p | flag.value
        else:
            self.clearFlag(flag)

    def clearFlag(self, flag: FlagBit | int | str) -> None:
        """Clear flag value

        :param flag: One of `N`, `V`, `B`, `D`, `I`, `Z`, `C`
        :type flag: FlagBit or int or str
        """
        if isinstance(flag, FlagBit):
            pass
        elif isinstance(flag, str):
            flag = FlagBit[flag]
        elif isinstance(flag, int):
            flag = FlagBit(flag)

        self.p = self.p & ~flag.value
        # self.p = self.p & (255 - flag)

    def clearFlags(self) -> None:
        """Clear all flag values
        """
        self.p = 0

    def ZN(self, v) -> None:
        """
        | The criteria for Z and N flags are standard.
        | Z gets set if the value is zero and
        | N gets set to the same value as bit 7 of the value.

        :param int v: Value
        """
        self.setFlag(FlagBit.Z, v == 0)
        self.setFlag(FlagBit.N, v & 0x80)

    @property
    def flags(self) -> str:
        return "{0:s}{1:s}{2:s}{3:s}{4:s}{5:s}{6:s}{7:s}".format(
            "N" if self.getFlag(FlagBit.N) else ".",
            "V" if self.getFlag(FlagBit.V) else ".",
            "U" if self.getFlag(FlagBit.U) else ".",
            "B" if self.getFlag(FlagBit.B) else ".",
            "D" if self.getFlag(FlagBit.D) else ".",
            "I" if self.getFlag(FlagBit.I) else ".",
            "Z" if self.getFlag(FlagBit.Z) else ".",
            "C" if self.getFlag(FlagBit.C) else ".",
        )

    def __repr__(self) -> str:
        return f"A: {self.a:0>2X} X: {self.x:0>2X} "\
               f"Y: {self.y:0>2X} S: {self.s:0>2X} "\
               f"PC: {self.pc:0>4X} P: {self.p:0>2X} [{self.flags}]"


class CPU:
    """The CPU object"""

    running: bool
    """CPU is running"""

    mmu: MMU
    """MMU"""

    op: Operation | None = None
    """Current Operation"""

    cc: int = 0
    """Holds the number of CPU cycles used during the last call to
    :py:meth:`py65emu.CPU.step`.
    Includes :py:attr:`py65emu.CPU.cc_extra`
    """

    cc_total: int = 0
    """
    Holds the total number of cycles during until
    :py:meth:`py65emu.CPU.reset`
    """

    cc_extra = 0
    """
    Holds the number of extra CPU cycles used during the last call to
    :py:meth:`py65emu.CPU.step`
    """

    def __init__(
        self,
        mmu: MMU,
        pc: int | None = None,
        stack_page: int = 0x1,
        magic: int = 0xEE,
        disable_bcd: bool = False,
        debug: bool = False,
    ):
        """
        Initialize CPU

        :param mmu: An instance of MMU
        :param pc: The starting address of the pc (program counter)
        :param stack_page: The index of the page which contains the stack. The
                           default for a 6502 is page 1 (the stack from
                           0x0100-0x1ff) but in some variants the stack page
                           may be elsewhere
        :param magic: A value needed for the illegal opcodes, XAA. This value
                      differs between different versions, even of the same
                      CPU. The default is 0xee
        :type mmu: MMU | None
        :type pc: int | None
        :type stack_page: int
        :type magic: int
        """
        self.mmu: MMU = mmu

        self.r: Registers = Registers()
        self.bcd_disabled = disable_bcd

        # Hold the number of CPU cycles used during the last call to
        # `self.step()`.
        # Since the CPU is in a "resetted" state, and a reset takes 7 cycles
        # we set the total cycles to 7
        self.cc: int = 0
        self.cc_extra = 0
        self.cc_total: int = 7

        # Which page the stack is in.  0x1 means that the stack is from
        # 0x100-0x1ff. In the 6502 this is always true but it's different
        # for other 65* varients.
        self.stack_page: int = stack_page

        self.magic: int = magic

        self.trigger_nmi: bool = False
        self.trigger_irq: bool = False
        self._previous_interrupt: bool = False
        self._interrupt: bool = False

        if pc:
            self.r.pc = pc
        else:
            # if pc is none get the address from $FFFD,$FFFC
            pass

        self.debug = debug
        self.opcodes = OpCodes(self)
        self.op = None

    def reset(self) -> None:
        """Reset everything (CPU, MMU, ...)"""
        self.r.reset(self.interrupts["RESET"])
        self.mmu.reset()
        self.trigger_nmi = False
        self.trigger_irq = False
        self._previous_interrupt = False
        self._interrupt = False
        self.op = None
        self.cc = 0
        self.cc_extra = 0
        self.cc_total = 7  # Reset takes 7 cycles

        self.running = True

    def step(self) -> None:
        """Execute the operation"""
        self.cc = 0
        self.cc_extra = 0

        opcode = self.nextByte()
        self._run_operation(opcode)

    def execute(self, instruction: list[int]) -> None:
        """
        Execute a single instruction independent of the program in memory.
        instruction is an array of bytes.

        :param list[int] instruction: List of unsigned integers to execute
        :rtype: None
        :return: None
        """
        for i in instruction:
            self.cc = 0
            self.cc_extra = 0

            self._run_operation(i)

    def _run_operation(self, opcode: int) -> None:
        """
        Execute a single instruction.

        :param list[int] instruction: List of unsigned integers to execute
        :rtype: None
        :return: None
        """
        self.op = self.opcodes[opcode]

        if self.debug is True:
            dasm = Disassembly(self.op, *self.op.get_operands())

        self.op.execute()

        self.cc += self.cc_extra

        if self.debug is True:
            print(
                f"{dasm!r: <44} {self.r!r: >44} "
                f"C: {self.cc:d} TC: {self.cc_total:d}"
            )
        self.cc_total += self.cc

        self.handle_interrupt()

    def handle_interrupt(self) -> None:
        """Handle interrupts (IRQ/NMI)"""
        if self._previous_interrupt:
            if self.trigger_nmi:
                self.process_nmi()
                self.trigger_nmi = False
            elif self.trigger_irq:
                self.process_irq()
                self.trigger_irq = False

    def readByte(self, addr: int) -> int:
        """
        Read byte from MMU

        :param int addr: 16 bit memory address
        :rtype: int
        :return: 8 bit (1 byte)
        """
        v = self.mmu.read(addr)
        self.increment_cycle_count()
        return v

    def readWord(self, addr: int) -> int:
        """
        Read 2 bytes (word) from MMU

        :param int addr: 16 bit memory address
        :rtype: int
        :return: 16 bit (2 byte)
        """
        low = self.readByte(addr)
        high = self.readByte(addr + 1)
        return (high << 8) + low

    def writeByte(self, addr: int, value: int) -> None:
        """
        Write byte to MMU

        :param int addr: 16 bit memory address
        :param int value: 8 bit value
        :rtype: None
        :return: None
        """
        self.mmu.write(addr, value)
        self.increment_cycle_count()

    def writeWord(self, addr: int, value: int) -> None:
        """
        Write 2 bytes (word) to MMU

        :param int addr: 16 bit memory address
        :param int value: 16 bit value
        :rtype: None
        :return: None
        """
        self.writeByte(addr, value & 0x00FF)
        self.writeByte(addr, (value >> 8) & 0x00FF)

    def nextByte(self) -> int:
        """
        Read next value (8 bit) from program counter

        :rtype: int
        :return: 8 bit
        """
        v = self.readByte(self.r.pc)
        self.r.pc = (self.r.pc + 1) & 0xFFFF
        return v

    def nextWord(self) -> int:
        """Read next value (16 bit) from program counter

        :rtype: int
        :return: 16 bit
        """
        low = self.nextByte()
        high = self.nextByte()
        return (high << 8) + low

    def stackPush(self, v: int) -> None:
        """Pushes value (8 bit) to stack

        :param int v: 8 bit
        """
        self.writeByte((self.stack_page * 0x100) + self.r.s, v)
        self.r.s = (self.r.s - 1) & 0xFF

    def stackPushWord(self, v: int) -> None:
        """Pushes value (16 bit) to stack

        :param int v: 16 bit
        """
        self.stackPush(v >> 8)
        self.stackPush(v & 0xFF)

    def stackPop(self) -> int:
        """Pops value (8 bit) from stack

        :rtype: int
        :return: 8 bit
        """
        v = self.readByte(self.stack_page * 0x100 + ((self.r.s + 1) & 0xFF))
        self.r.s = (self.r.s + 1) & 0xFF
        return v

    def stackPopWord(self) -> int:
        """Pops value (16 bit) from stack

        :rtype: int
        :return: 16 bit
        """
        return self.stackPop() + (self.stackPop() << 8)

    def fromBCD(self, v) -> int:
        """Converts BCD (Binary Coded Decimal) to int

        :param int v: BCD coded value
        :rtype: int
        :return: 8 bit
        """
        return (((v & 0xF0) // 0x10) * 10) + (v & 0xF)

    def toBCD(self, v) -> int:
        """
        Converts int to BCD (Binary Coded Decimal)

        :param int v: Integer
        :rtype: int
        :return: BCD coded value

        """
        return int(math.floor(v / 10)) * 16 + (v % 10)

    def signedHex(self, v: int, bits: int = 8) -> int:
        """
        Converts unsigned int value to signed int value

        :param int v: The unsigned int value
        :param int bits: | Number of bits the value has. (Default 8)
                         | 8 bits   = byte
                         | 16 bits  = word
                         | 32 bits  = long
                         | 64 bits  = long long
                         | 128 bits = double
        :rtype: int
        :return: Signed integer

        """
        if v >= (1 << (bits - 1)):
            v -= (1 << bits)
        return v

    def fromTwosCom(self, v: int) -> int:
        """Converts unsigned int value to signed int value

        :param int v: The unsigned int value
        :rtype: int
        :return: Signed integer
        """
        return (v & 0x7F) - (v & 0x80)

    def increment_cycle_count(self, cycles: int = 1) -> None:
        """Increment cycle on current operation

        :param int cycles: Number of cycles to increment with. (Default: 1)
        """
        self.cc = (self.cc + cycles) & 0xFF
        self._previousInterrupt = self._interrupt
        self._interrupt = (
            self.trigger_nmi or
            (self.trigger_irq and self.r.getFlag(FlagBit.I) is False)
        )

    def increment_extra_cycle(self, cycles: int = 1) -> None:
        """Add extra cycle to current operation"""
        self.cc_extra = (self.cc_extra + cycles) & 0xFF

    interrupts = {
        "ABORT": 0xfff8,
        "COP": 0xfff4,
        "IRQ": 0xfffe,
        "BRK": 0xfffe,
        "NMI": 0xfffa,
        "RESET": 0xfffc
    }
    """Dictionary with interrupt types and their corresponding addresses"""

    def interruptAddress(self, irq_type: str) -> int:
        """
        Get values from interrupt vector

        .. seealso::
           :py:attr:`.interrupts`


        :param str irq_type: Interrupt type, key
                             from :py:attr:`.interrupts`
        """
        return self.mmu.readWord(self.interrupts[irq_type])

    def interruptRequest(self) -> None:
        """Trigger interrupt on next operation"""
        self.trigger_irq = True

    def breakOperation(self, irq_type: str) -> None:
        """
        The BRK routine. Called when a BRK occurs.
        Also called from NMI/IRQ operations

        :param str irq_type: Interrupt type, key
                             from :py:attr:`py65emu.CPU.interrupts`
        """
        self.increment_cycle_count()
        self.stackPushWord(self.r.pc + 1)
        self.increment_cycle_count()

        if irq_type == "BRK":
            self.stackPush(self.r.p | FlagBit.B.value)
        else:
            self.stackPush(self.r.p)

        self.increment_cycle_count()

        self.r.setFlag("I")
        self.r.pc = self.interruptAddress(irq_type)

    def process_nmi(self) -> None:
        """
        Proccesses NMI Interrupt.

        .. seealso::
           :py:meth:`.breakOperation`
        """
        self.r.pc -= 1
        self.breakOperation("NMI")

    def process_irq(self) -> None:
        """
        Proccesses IRQ Interrupt.

        .. seealso::
           :py:meth:`.breakOperation`
        """
        if self.r.getFlag(FlagBit.I):
            return None
        self.r.pc -= 1
        self.breakOperation("IRQ")

    # region Addressing modes
    def im_a(self) -> int:
        """
        Immediate - OPC #$BB (Also: Accumulator - OPC A, Implied - OPC)

        +--------------+--------------------------------------------------+
        | Address Mode | Description                                      |
        +==============+==================================================+
        | Immediate    | The instruction expects the next byte to be used |
        |              | as a value, so we'll prep the read address to    |
        |              | point to the next byte                           |
        +--------------+--------------------------------------------------+
        | Method is also used for following address modes                 |
        +--------------+--------------------------------------------------+
        | Accumulator  | Operand is always AC                             |
        +--------------+--------------------------------------------------+
        | Implied      | There is no additional data required for this    |
        |              | instruction. The instruction does something very |
        |              | simple like like sets a status bit. However, we  |
        |              | will target the accumulator, for instructions    |
        |              | like PHA                                         |
        +--------------+--------------------------------------------------+

        :rtype: int
        :return: 8 bit value
        """
        return self.nextByte()

    def z_a(self) -> int:
        """
        Zero Page - OPC $LL

        +--------------+--------------------------------------------------+
        | Address Mode | Description                                      |
        +--------------+--------------------------------------------------+
        | Zero Page    | To save program bytes, zero page addressing      |
        |              | allows you to absolutely address a location in   |
        |              | first 0xFF bytes of address range. Clearly this  |
        |              | only requires one byte instead of the usual two. |
        +--------------+--------------------------------------------------+

        :rtype: int
        :return: 8 bit address
        """
        return self.nextByte() & 0xFF

    def zx_a(self) -> int:
        """
        Zero Page with X Offset - OPC $LL, X

        +-------------------------+---------------------------------------+
        | Address Mode            | Description                           |
        +-------------------------+---------------------------------------+
        | Zero Page with X Offset | Fundamentally the same as Zero Page   |
        |                         | addressing, but the contents of the   |
        |                         | X Register is added to the supplied   |
        |                         | single byte address. This is useful   |
        |                         | for iterating through ranges within   |
        |                         | the first page.                       |
        +-------------------------+---------------------------------------+

        :rtype: int
        :return: 8 bit address
        """
        self.increment_cycle_count()
        return (self.nextByte() + self.r.x) & 0xFF

    def zy_a(self) -> int:
        """
        Zero Page with Y Offset - OPC $LL, Y

        +-------------------------+---------------------------------------+
        | Address Mode            | Description                           |
        +-------------------------+---------------------------------------+
        | Zero Page with Y Offset | Same as above but uses Y Register for |
        |                         | offset                                |
        +-------------------------+---------------------------------------+

        :rtype: int
        :return: 8 bit address
        """
        self.increment_cycle_count()
        return (self.nextByte() + self.r.y) & 0xFF

    def rel_a(self) -> int:
        """
        Relative - OPC $BB (This address mode is exclusive to branch instr.)

        +--------------+--------------------------------------------------+
        | Address Mode | Description                                      |
        +--------------+--------------------------------------------------+
        | Relative     | This address mode is exclusive to branch         |
        |              | instructions. The address must reside within     |
        |              | -128 to +127 of the branch instruction, i.e. you |
        |              | cant directly branch to any address in the       |
        |              | addressable range.                               |
        +--------------+--------------------------------------------------+

        :rtype: int
        :return: 16 bit address
        """
        d = self.nextByte()
        return (self.r.pc + self.fromTwosCom(d)) & 0xFFFF

    def a_a(self) -> int:
        """
        Absolute - OPC $LLHH

        +--------------+--------------------------------------------------+
        | Address Mode | Description                                      |
        +--------------+--------------------------------------------------+
        | Absolute     | A full 16-bit address is loaded and used         |
        +--------------+--------------------------------------------------+

        :rtype: int
        :return: 16 bit address
        """
        return self.nextWord()

    def ax_a(self) -> int:
        """
        Absolute with X Offset - OPC $LLHH, X

        +------------------------+----------------------------------------+
        | Address Mode           | Description                            |
        +------------------------+----------------------------------------+
        | Absolute with X Offset | Fundamentally the same as absolute     |
        |                        | addressing, but the contents of the    |
        |                        | X Register is added to the supplied    |
        |                        | two byte address. If the resulting     |
        |                        | address changes the page, an           |
        |                        | additional clock cycle is required     |
        +------------------------+----------------------------------------+

        :rtype: int
        :return: 16 bit address
        """
        o = self.nextWord()
        a = o + self.r.x

        special_op = [
            0x1E, 0x1F, 0x3E, 0x3F, 0x5E, 0x5F, 0x7E,
            0x7F, 0x9D, 0x9F, 0xDE, 0xDF, 0xFE, 0xFF
        ]
        if o & 0xFF00 != a & 0xFF00 and (
            self.op and self.op.opcode not in special_op
        ):
            # self.cc += 1
            self.increment_extra_cycle()
        elif self.op and self.op.opcode in special_op:
            self.increment_cycle_count()

        return a & 0xFFFF

    def ay_a(self) -> int:
        """
        Absolute with Y Offset - OPC $LLHH, Y

        +------------------------+----------------------------------------+
        | Address Mode           | Description                            |
        +------------------------+----------------------------------------+
        | Absolute with Y Offset | Fundamentally the same as absolute     |
        |                        | addressing, but the contents of the    |
        |                        | Y Register is added to the supplied    |
        |                        | two byte address. If the resulting     |
        |                        | address changes the page, an           |
        |                        | additional clock cycle is required     |
        +------------------------+----------------------------------------+

        :rtype: int
        :return: 16 bit address
        """
        o = self.nextWord()
        a = o + self.r.y

        special_op = [0x1B, 0x3B, 0x5B, 0x7B, 0x99, 0xDB, 0xFB]
        if o & 0xFF00 != a & 0xFF00 and (
            self.op and self.op.opcode not in special_op
        ):
            # self.cc += 1
            self.increment_extra_cycle()
        elif self.op and self.op.opcode in special_op:
            self.increment_cycle_count()

        return a & 0xFFFF

    def i_a(self) -> int:
        """
        Indirect - OPC ($LLHH) (Only used by indirect JMP)

        +--------------+--------------------------------------------------+
        | Address Mode | Description                                      |
        +--------------+--------------------------------------------------+
        | Indirect     | The supplied 16-bit address is read to get the   |
        |              | actual 16-bit address. This is instruction is    |
        |              | unusual in that it has a bug in the hardware!    |
        |              | To emulate its function accurately, we also need |
        |              | to emulate this bug. If the low byte of the      |
        |              | supplied address is 0xFF, then to read the high  |
        |              | byte of the actual address we need to cross a    |
        |              | page boundary. This doesn't actually work on the |
        |              | chip as designed, instead it wraps back around   |
        |              | in the same page, yielding an invalid actual     |
        |              | address                                          |
        +--------------+--------------------------------------------------+


        :rtype: int
        :return: 16 bit address
        """
        i = self.nextWord()
        # Doesn't carry, so if the low byte is in the XXFF position
        # Then the high byte will be XX00 rather than XY00
        if i & 0xFF == 0xFF:
            j = i - 0xFF
        else:
            j = i + 1

        return ((self.readByte(j) << 8) + self.readByte(i)) & 0xFFFF

    def ix_a(self) -> int:
        """
        Indirect X - OPC ($LL, X)

        +--------------+--------------------------------------------------+
        | Address Mode | Description                                      |
        +--------------+--------------------------------------------------+
        | Indirect X   | The supplied 8-bit address is offset by          |
        |              | X Register to index a location in page 0x00. The |
        |              | actual 16-bit address is read from this location |
        +--------------+--------------------------------------------------+

        :rtype: int
        :return: 16 bit address
        """
        # self.increment_extra_cycle()
        self.increment_cycle_count()
        i = (self.nextByte() + self.r.x) & 0xFF
        return (
            ((self.readByte((i + 1) & 0xFF) << 8) + self.readByte(i)) & 0xFFFF
        )

    def iy_a(self) -> int:
        """
        Indirect Y - OPC ($LL), Y

        +--------------+--------------------------------------------------+
        | Address Mode | Description                                      |
        +--------------+--------------------------------------------------+
        | Indirect Y   | The supplied 8-bit address is offset by          |
        |              | X Register to index a location in page 0x00. The |
        |              | actual 16-bit address is read, and the contents  |
        |              | of Y Register is added to it to offset it. If    |
        |              | the offset causes change in page then an         |
        |              | additional clock cycle is required.              |
        +--------------+--------------------------------------------------+

        :rtype: int
        :return: 16 bit address
        """
        i = self.nextByte()
        o = (self.readByte((i + 1) & 0xFF) << 8) + self.readByte(i)
        a = o + self.r.y

        special_op = [0x13, 0x33, 0x53, 0x73, 0x91, 0xD3, 0xF3]
        if o & 0xFF00 != a & 0xFF00 and (
            self.op and self.op.opcode not in special_op
        ):
            # self.cc += 1
            self.increment_extra_cycle()
        elif self.op and self.op.opcode in special_op:
            self.increment_cycle_count()

        return a & 0xFFFF

    # endregion Addressing modes

    # region Return values based on the addressing mode
    def im(self) -> int:
        """
        Alias method for :py:meth`py65emu.cpu.CPU.im_a`

        :return: 8 bit value
        :rtype: int
        """
        return self.im_a()

    def z(self) -> int:
        """
        Return memory value from Zero Page

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.z_a())

    def zx(self) -> int:
        """
        Return memory value from Zero Page, X

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.zx_a())

    def zy(self) -> int:
        """
        Return memory value from Zero Page, Y

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.zy_a())

    def a(self) -> int:
        """
        Return memory value from Absolute

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.a_a())

    def ax(self) -> int:
        """
        Return memory value from Absolute, X

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.ax_a())

    def ay(self) -> int:
        """
        Return memory value from Absolute, Y

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.ay_a())

    def i(self) -> int:
        """
        Return memory value from Indirect

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.i_a())

    def ix(self) -> int:
        """
        Return memory value from Indirect, X

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.ix_a())

    def iy(self) -> int:
        """
        Return memory value from Indirect, Y

        :return: 8 bit value
        :rtype: int
        """
        return self.readByte(self.iy_a())

    # endregion Return values based on the addressing mode

    # region Instructions Mnemonics

    def ADC(self, v2: int) -> None:
        """
        Add with Carry In

        +---------------+---------------+
        | Function      | `N Z C I D V` |
        +===============+===============+
        | A = A + M + C | `N Z C - - -` |
        +---------------+---------------+

        :param int v2: 8 bit value
        """
        v1 = self.r.a

        if self.r.getFlag("D") and self.bcd_disabled is False:  # decimal mode
            d1 = self.fromBCD(v1)
            d2 = self.fromBCD(v2)
            r = d1 + d2 + self.r.getFlag("C")
            self.r.a = self.toBCD(r % 100)

            self.r.setFlag("C", r > 99)
        else:
            r = v1 + v2 + self.r.getFlag("C")
            self.r.a = r & 0xFF

            self.r.setFlag("C", r > 0xFF)

        self.r.ZN(self.r.a)
        self.r.setFlag("V", ((~(v1 ^ v2)) & (v1 ^ r) & 0x80))

    def AND(self, v: int) -> None:
        """
        Bitwise Logic AND

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | A = A & M | `N Z - - - -` |
        +-----------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = (self.r.a & v) & 0xFF
        self.r.ZN(self.r.a)

    def ASL(self, a: str | int) -> None:
        """
        Arithmetic Shift Left

        +------------------------+---------------+
        | Function               | `N Z C I D V` |
        +========================+===============+
        | A = C <- (A << 1) <- 0 | `N Z - - - V` |
        +------------------------+---------------+

        :param a: 16 bit address, or accumulator
        :type a: int | str
        """
        if isinstance(a, str):
            v = self.r.a
            self.increment_cycle_count()
        else:
            v = self.readByte(a)
            self.writeByte(a, v)

        v = v << 1
        self.r.setFlag("C", v > 0xFF)
        self.r.ZN(v & 0xFF)

        if isinstance(a, str):
            self.r.a = v & 0xFF
        else:
            self.writeByte(a, v)

    def BIT(self, v: int) -> None:
        """
        Test Bits in Memory with Accumulator

        +----------+---------------+
        | Function | `N Z C I D V` |
        +==========+===============+
        | A & M    | `N Z - - - V` |
        +----------+               |
        | M6 -> V  |               |
        +----------+               |
        | M7 -> N  |               |
        +----------+---------------+

        :param int v: 8 bit value
        """
        self.r.setFlag("Z", self.r.a & v == 0)
        self.r.setFlag("N", v & 0x80)
        self.r.setFlag("V", v & 0x40)

    def B(self, v: tuple[str, bool]) -> None:
        """
        Aggregated method for all "Branch On"-operations
        For instance, BCC (Branch Carry Clear) will call B(('C', False))

        +------+------+---------------------------+----------+---------------+
        | Name | OPC  | Instruction               | Function | `N Z C I D V` |
        +======+======+===========================+==========+===============+
        | BCC  | 0x90 | Branch on Carry Clear     | C = 0    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+
        | BCS  | 0xB0 | Branch on Carry Set       | C = 1    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+
        | BEQ  | 0xF0 | Branch on Result Zero     | Z = 1    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+
        | BNE  | 0xD0 | Branch on Result not Zero | Z = 0    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+
        | BMI  | 0x30 | Branch on Result Minus    | N = 1    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+
        | BPL  | 0x10 | Branch on Result Plus     | N = 0    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+
        | BVC  | 0x50 | Branch on Overflow Clear  | V = 0    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+
        | BVS  | 0x70 | Branch on Overflow Set    | V = 1    | `- - - - - -` |
        +------+------+---------------------------+----------+---------------+

        :param v: v is a tuple of (flag, boolean)
        :type v: tuple[str, bool]
        """

        pc_rel = self.rel_a()
        if self.r.getFlag(v[0]) is v[1]:
            self.increment_extra_cycle()

            if pc_rel & 0xFF00 != self.r.pc & 0xFF00:
                self.increment_extra_cycle()
            self.r.pc = pc_rel

    def BRK(self, _: int) -> None:
        """
        Break

        +--------------+---------------------------+---------------+
        | Instruction  | Function                  | `N Z C I D V` |
        +==============+===========================+===============+
        | Break        | Program Sourced Interrupt | `- - - - - -` |
        +--------------+---------------------------+---------------+

        :param int _: Ignored
        """
        self.breakOperation("BRK")

    def CP(self, r: int, v: int) -> None:
        """
        Support method for CMP/CPX/CPY methods

        :param int r: 8 bit register value
        :param int v: 8 bit value
        """
        o = (r - v) & 0xFF
        self.r.setFlag("Z", (o & 0xFF) == 0)
        self.r.setFlag("C", r >= v)
        self.r.setFlag("N", o & 0x80)

    def CMP(self, v: int) -> None:
        """
        Compare Accumulator

        +---------------------+-------------------+---------------+
        | Instruction         | Function          | `N Z C I D V` |
        +=====================+===================+===============+
        | Compare Accumulator | C <- A >= M       | `N Z C - - -` |
        |                     +-------------------+               |
        |                     | Z <- (A - M) == 0 |               |
        +---------------------+-------------------+---------------+

        :param int v: 8 bit value
        """
        self.CP(self.r.a, v)

    def CPX(self, v: int) -> None:
        """
        Compare X Register

        +-------------------+---------------+
        | Function          | `N Z C I D V` |
        +===================+===============+
        | C <- X >= M       | `N Z C - - -` |
        +-------------------+               |
        | Z <- (X - M) == 0 |               |
        +-------------------+---------------+

        :param int v2: 8 bit value
        """
        self.CP(self.r.x, v)

    def CPY(self, v: int) -> None:
        """
        Compare Y Register

        +-------------------+---------------+
        | Function          | `N Z C I D V` |
        +===================+===============+
        | C <- Y >= M       | `N Z C - - -` |
        +-------------------+               |
        | Z <- (Y - M) == 0 |               |
        +-------------------+---------------+

        :param int v: 8 bit value
        """
        self.CP(self.r.y, v)

    def DEC(self, a: int) -> None:
        """
        Decrement Value at Memory Location

        +------------------------------------+-----------+---------------+
        | Instruction                        | Function  | `N Z C I D V` |
        +====================================+===========+===============+
        | Decrement Value at Memory Location | M = M - 1 | `N Z - - - -` |
        +------------------------------------+-----------+---------------+

        :param int a: 16 bit address location
        """
        v = self.readByte(a)
        self.writeByte(a, v & 0xFF)
        v = (v - 1) & 0xFF
        self.writeByte(a, v)
        self.r.ZN(v)

    def DEX(self, _) -> None:
        """
        Decrement X Register

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | X = X - 1 | `N Z - - - -` |
        +-----------+---------------+

        :param int _: Ignored
        """
        self.r.x = (self.r.x - 1) & 0xFF
        self.r.ZN(self.r.x)
        self.increment_cycle_count()

    def DEY(self, _) -> None:
        """
        Decrement Y Register

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | Y = Y - 1 | `N Z - - - -` |
        +-----------+---------------+

        :param int _: Ignored
        """
        self.r.y = (self.r.y - 1) & 0xFF
        self.r.ZN(self.r.y)
        self.increment_cycle_count()

    def EOR(self, v: int) -> None:
        """
        Bitwise Logic XOR

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | A = A ^ M | `N Z - - - -` |
        +-----------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = self.r.a ^ v
        self.r.ZN(self.r.a)

    """Flag Instructions."""

    def SE(self, v: str) -> None:
        """
        Aggregated method for all "Set Flag"-operations

        +------+------+------------------------+-----------+---------------+
        | Name | OPC  | Instruction            | Function  | `N Z C I D V` |
        +======+======+========================+===========+===============+
        | SEC  | 0x38 | Set Carry Flag         | 1 -> C    | `- - 1 - - -` |
        +------+------+------------------------+-----------+---------------+
        | SED  | 0xF8 | Set Decimal Flag       | 1 -> D    | `- - - - 1 -` |
        +------+------+------------------------+-----------+---------------+
        | SEI  | 0x78 | Set IRQ Disable Status | 1 -> I    | `- - - 1 - -` |
        +------+------+------------------------+-----------+---------------+

        :param v: One of `D`, `I`, `C`
        :type v: FlagBit or int or str
        """
        self.r.setFlag(v)
        self.increment_cycle_count()

    def CL(self, v: str) -> None:
        """
        Aggregated method for all "Clear Flag"-operations

        +------+------+--------------------------+-----------+---------------+
        | Name | OPC  | Instruction              | Function  | `N Z C I D V` |
        +======+======+==========================+===========+===============+
        | CLC  | 0x18 | Clear Carry Flag         | 0 -> C    | `- - 0 - - -` |
        +------+------+--------------------------+-----------+---------------+
        | CLD  | 0xD8 | Clear Decimal Flag       | 0 -> D    | `- - - - 0 -` |
        +------+------+--------------------------+-----------+---------------+
        | CLI  | 0x58 | Clear IRQ Disable Status | 0 -> I    | `- - - 0 - -` |
        +------+------+--------------------------+-----------+---------------+
        | CLV  | 0xB8 | Clear Overflow Flag      | 0 -> V    | `- - - - - 0` |
        +------+------+--------------------------+-----------+---------------+

        :param v: One of `V`, `D`, `I`, `C`
        :type v: FlagBit or int or str
        """
        self.r.clearFlag(v)
        self.increment_cycle_count()

    def INC(self, a: int) -> None:
        """
        Increment Value at Memory Location

        +------------------------------------+-----------+---------------+
        | Instruction                        | Function  | `N Z C I D V` |
        +====================================+===========+===============+
        | Increment Value at Memory Location | M = M + 1 | `N Z - - - -` |
        +------------------------------------+-----------+---------------+

        :param int a: 16 bit address location
        """
        v = self.readByte(a)
        self.writeByte(a, v & 0xFF)
        v = (v + 1) & 0xFF
        self.writeByte(a, v)
        self.r.ZN(v)

    def INX(self, _) -> None:
        """
        Increment X Register

        +----------------------+-----------+---------------+
        | Instruction          | Function  | `N Z C I D V` |
        +======================+===========+===============+
        | Increment X Register | X = X + 1 | `N Z - - - -` |
        +----------------------+-----------+---------------+

        :param int _: Ignored
        """
        self.r.x = (self.r.x + 1) & 0xFF
        self.r.ZN(self.r.x)
        self.increment_cycle_count()

    def INY(self, _) -> None:
        """
        Increment Y Register

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | Y = Y + 1 | `N Z - - - -` |
        +-----------+---------------+

        :param int _: Ignored
        """
        self.r.y = (self.r.y + 1) & 0xFF
        self.r.ZN(self.r.y)
        self.increment_cycle_count()

    def JMP(self, a: int) -> None:
        """
        Jump to new Location

        +-----------------+---------------+
        | Function        | `N Z C I D V` |
        +=================+===============+
        | (PC + 1) -> PCL | `- - - - - -` |
        +-----------------+               |
        | (PC + 2) -> PCH |               |
        +-----------------+---------------+

        :param int a: Address to jump new PC
        """
        self.r.pc = a

    def JSR(self, a: int) -> None:
        """
        Jump To Sub-Routine

        +-----------------+---------------+
        | Function        | `N Z C I D V` |
        +=================+===============+
        | push (PC + 2)   | `- - - - - -` |
        +-----------------+               |
        | (PC + 1) -> PCL |               |
        +-----------------+               |
        | (PC + 2) -> PCH |               |
        +-----------------+---------------+

        :param int a: New PC Value
        """
        self.stackPushWord(self.r.pc - 1)
        self.r.pc = a
        self.increment_cycle_count()

    def LDA(self, v: int) -> None:
        """
        Load The Accumulator

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | A = M     | `N Z - - - -` |
        +-----------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = v
        self.r.ZN(self.r.a)

    def LDX(self, v: int) -> None:
        """
        Load X Register

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | X = M     | `N Z - - - -` |
        +-----------+---------------+

        :param int v: 8 bit value
        """
        self.r.x = v
        self.r.ZN(self.r.x)

    def LDY(self, v: int) -> None:
        """
        Load Y Register

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | Y = M     | `N Z - - - -` |
        +-----------+---------------+

        :param int v: 8 bit value
        """
        self.r.y = v
        self.r.ZN(self.r.y)

    def LSR(self, a: str | int) -> None:
        """
        Logical Shift Right

        +------------------------+---------------+
        | Function               | `N Z C I D V` |
        +========================+===============+
        | A = C <- (A << 1) <- 0 | `0 Z C - - -` |
        +------------------------+---------------+

        :param a: 16 bit address, or accumulator
        :type a: int | str
        """
        if isinstance(a, str):
            v = self.r.a
            self.increment_cycle_count()
        else:
            v = self.readByte(a)
            self.writeByte(a, v)

        self.r.setFlag("C", v & 0x01)
        v = v >> 1
        self.r.ZN(v)

        if isinstance(a, str):
            self.r.a = v
        else:
            self.writeByte(a, v)

    def NOP(self, _) -> None:
        """
        No Operation

        Sadly not all NOPs are equal, Ive added a few here
        based on `CPU_unofficial_opcodes`_ and will add more
        based on game compatibility, and ultimately I'd like
        to cover all illegal opcodes too

        ::
            **NOP (`$1A`, `$3A`, `$5A`, `$7A`, `$DA`, `$EA`, `$FA`)**

            Address Mode: Implied

            The official NOP (`$EA`) and six unofficial NOPs do nothing.

            **SKB - OPC (`$80`, `$82`, `$89`, `$C2`, `$E2`)**

            Address Mode: Immediate
            `$89` affects NVZ flags, like :py:meth:`.BIT`

            **IGN - OPC (`$0C`)**
            Address Mode: Absolute

            **IGN - OPC (`$1C`, `$3C`, `$5C`, `$7C`, `$DC`, `$FC`)
            Address Mode: Absolute, X

            **IGN - OPC (`$04`, `$44`, `$64`)
            Address Mode: Zero Page

            **IGN - OPC (`$14`, `$34`, `$54`, `$74`, `$D4`, `$F4`)
            Address Mode: Zero Page, X

        .. _CPU_unofficial_opcodes:
            https://www.nesdev.org/wiki/Programming_with_unofficial_opcodes

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        |           | `- - - - - -` |
        +-----------+---------------+

        :param int _: Ignored
        """
        if self.op and self.op.mode not in ['im']:
            self.increment_cycle_count()

    def ORA(self, v: int) -> None:
        """
        Bitwise Logic OR

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | A = A | M | `N Z - - - -` |
        +-----------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = self.r.a | v
        self.r.ZN(self.r.a)

    def P(self, v: tuple[str, str]) -> None:
        """
        Aggregated method for all Stack operations, PusH and PulL.

        v is a tuple where the first value is either PH or PL, specifying the
        action and the second is the source or target register, either A or P,
        meaning the Accumulator or the Processor status flag.

        +------+------+------------------+----------+---------------+
        | Name | OPC  | Instruction      | Function | `N Z C I D V` |
        +======+======+==================+==========+===============+
        | PHA  | 0x48 | Push Accumulator | push A   | `- - - - - -` |
        |      |      | on Stack         |          |               |
        +------+------+------------------+----------+---------------+
        | PHP  | 0x08 | Push Register P  | push SR  | `- - - - - -` |
        |      |      | on Stack         |          |               |
        +------+------+------------------+----------+---------------+
        | PLA  | 0x68 | Pull Accumulator | pull A   | `N Z - - - -` |
        |      |      | from Stack       |          |               |
        +------+------+------------------+----------+---------------+
        | PLP  | 0x28 | Pull Register P  | pull SR  | from    stack |
        |      |      | from Stack       |          |               |
        +------+------+------------------+----------+---------------+

        :param a: PHA would be `P((PH, A))`, and
                  PLP `P((PL, P))`
        :type a: tuple[str, str]
        """
        a, r = v

        self.increment_cycle_count()
        if a == "PH":
            register = getattr(self.r, r)
            if r == "p":
                register |= 0b00110000

            self.stackPush(register)
        else:
            setattr(self.r, r, self.stackPop())

            if r == "a":
                self.r.ZN(self.r.a)
            elif r == "p":
                self.r.clearFlag(FlagBit.B)
                self.r.setFlag(FlagBit.U)
                # self.r.p = self.r.p | 0b00100000

            self.increment_cycle_count()

    def ROL(self, a: str | int) -> None:
        """
        Rotate Left

        +----------+---------------+
        | Function | `N Z C I D V` |
        +==========+===============+
        | (C << 1) | `N Z C - - -` |
        +----------+---------------+

        :param a: 16 bit address, or accumulator
        :type a: int | str
        """
        if isinstance(a, str):
            v_old = self.r.a
            self.increment_cycle_count()
        else:
            v_old = self.readByte(a)
            self.writeByte(a, v_old)

        v_new = ((v_old << 1) + self.r.getFlag("C")) & 0xFF
        self.r.setFlag("C", v_old & 0x80)
        self.r.ZN(v_new)

        if isinstance(a, str):
            self.r.a = v_new
        else:
            self.writeByte(a, v_new)

    def ROR(self, a: str | int) -> None:
        """
        Rotate Right

        +----------+---------------+
        | Function | `N Z C I D V` |
        +==========+===============+
        | (C >> 1) | `N Z C - - -` |
        +----------+---------------+

        :param a: 16 bit address, or accumulator
        :type a: int | str
        """
        if isinstance(a, str):
            v_old = self.r.a
            self.increment_cycle_count()
        else:
            v_old = self.readByte(a)
            self.writeByte(a, v_old)

        v_new = ((v_old >> 1) + self.r.getFlag("C") * 0x80) & 0xFF
        self.r.setFlag("C", v_old & 0x01)
        self.r.ZN(v_new)
        if isinstance(a, str):
            self.r.a = v_new
        else:
            self.writeByte(a, v_new)

    def RTI(self, _) -> None:
        """
        Return from Interrupt

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | pull SR   | `from  stack` |
        +-----------+               |
        | pull PC   |               |
        +-----------+---------------+

        :param int _: Ignored
        """
        self.r.p = self.stackPop()
        self.r.setFlag(FlagBit.U)
        self.increment_cycle_count()

        self.r.pc = self.stackPopWord()
        self.increment_cycle_count()

    def RTS(self, _) -> None:
        """
        Return from Subroutine

        +--------------+---------------+
        | Function     | `N Z C I D V` |
        +==============+===============+
        | pull PC      | `- - - - - -` |
        +--------------+               |
        | PC + 1 -> PC |               |
        +--------------+---------------+

        :param int _: Ignored
        """
        self.increment_cycle_count(3)
        self.r.pc = (self.stackPopWord() + 1) & 0xFFFF

    def SBC(self, v2: int) -> None:
        """
        Subtraction with Borrow In

        +---------------------+---------------+
        | Function            | `N Z C I D V` |
        +=====================+===============+
        | A = A - M - (1 - C) | `N Z C - - V` |
        +---------------------+---------------+

        :param int v2: 8 bit value
        """
        v1 = self.r.a
        if self.r.getFlag("D") and self.bcd_disabled is False:
            d1 = self.fromBCD(v1)
            d2 = self.fromBCD(v2)
            r = d1 - d2 - (not self.r.getFlag("C"))
            self.r.a = self.toBCD(r % 100)
        else:
            """
            r = v1 + (v2 ^ 0xFF) + self.r.getFlag("C")

            self.r.a = r & 0xFF
            """
            r = v1 - v2 - (not self.r.getFlag('C'))
            self.r.a = r & 0xff

        self.r.setFlag("C", r >= 0)
        self.r.setFlag("V", ((v1 ^ v2) & (v1 ^ r) & 0x80))
        self.r.ZN(self.r.a)

    def STA(self, a: int) -> None:
        """
        Store Accumulator at Address

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | M = A     | `- - - - - -` |
        +-----------+---------------+

        :param int a: 16 bit address
        """
        self.writeByte(a, self.r.a)

    def STX(self, a: int) -> None:
        """
        Store X Register at Address

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | M = X     | `- - - - - -` |
        +-----------+---------------+

        :param int a: 16 bit address
        """
        self.writeByte(a, self.r.x)

    def STY(self, a: int) -> None:
        """
        Store Y Register at Address

        +-----------+---------------+
        | Function  | `N Z C I D V` |
        +===========+===============+
        | M = Y     | `- - - - - -` |
        +-----------+---------------+

        :param int a: 16 bit address
        """
        self.writeByte(a, self.r.y)

    def T(self, a: tuple[str, str]) -> None:
        """
        Aggregated method for all transfer register to register operations.

        Parameter is a 2-tuple with source and destination register,
        so a call to TAX-operation (`$AA`) would be:

        >>> cpu.T(('a', 'x'))

        +--------------------------------------+----------+---------------+
        | Instruction                          | Function | `N Z C I D V` |
        +======================================+==========+===============+
        | Transfer Accumulator to Register X   | A -> X   | `N Z - - - -` |
        +--------------------------------------+----------+---------------+
        | Transfer Accumulator to Register Y   | A -> Y   | `N Z - - - -` |
        +--------------------------------------+----------+---------------+
        | Transfer Stack Pointer to Register X | SP -> X  | `N Z - - - -` |
        +--------------------------------------+----------+---------------+
        | Transfer Register X to Accumulator   | X -> A   | `N Z - - - -` |
        +--------------------------------------+----------+---------------+
        | Transfer Register X to Stack Pointer | X -> SP  | `- - - - - -` |
        +--------------------------------------+----------+---------------+
        | Transfer Register Y to Accumulator   | Y -> A   | `N Z - - - -` |
        +--------------------------------------+----------+---------------+

        :param a: tuple with (source, destination)
        :type a: tuple[str, str]
        """
        s, d = a
        self.increment_cycle_count()
        setattr(self.r, d, getattr(self.r, s))
        if d != "s":
            self.r.ZN(getattr(self.r, d))

    # endregion Instructions Mnemonics

    # region Illegal/Undocumented Operations

    """
    Illegal Opcodes
    ---------------

    Opcodes which were not officially documented but still have
    and effect.  The behavior for each of these is based on the following:

    -http://www.ataripreservation.org/websites/freddy.offenga/illopc31.txt

    -http://wiki.nesdev.com/w/index.php/Programming_with_unofficial_opcodes

    -www.ffd2.com/fridge/docs/6502-NMOS.extra.opcodes

    The behavior is not consistent across the various resources so I don't
    promise 100% hardware accuracy here.

    Other names for the opcode are in comments on the function defintion
    line.
    """

    def AAC(self, v: int) -> None:  # ANC
        """
        AND oper + set C as ASL

        .. important:: Illegal Opcode

        +-------------+---------------+
        | Function    | `N Z C I D V` |
        +=============+===============+
        | A AND oper  | `N Z C - - -` |
        +-------------+               |
        | bit(7) -> C |               |
        +-------------+---------------+

        :param int v: 8 bit value
        """
        self.AND(v)
        self.r.setFlag("C", self.r.getFlag("N"))

    def AAX(self, a: int) -> None:  # SAX, AXS
        """
        A and X are put on the bus at the same time (resulting effectively in
        an AND operation) and stored in M

        .. important:: Illegal Opcode

        +--------------+---------------+
        | Function     | `N Z C I D V` |
        +==============+===============+
        | A AND X -> M | `- - - - - -` |
        +--------------+---------------+

        :param int a: 16 bit address
        """
        r = self.r.a & self.r.x
        self.writeByte(a, r)
        # There is conflicting information whether this effects P.
        # self.r.ZN(r)

    def ARR(self, v: int) -> None:
        """
        AND oper + ROR

        .. important:: Illegal Opcode

        +----------------------+---------------+
        | Function             | `N Z C I D V` |
        +======================+===============+
        | A AND oper           | `N Z C - - V` |
        +----------------------+               |
        | C -> [76543210] -> C |               |
        +----------------------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = (self.r.getFlag('C') << 7) | (((self.r.a & v) & 0xFF) >> 1)
        self.r.setFlag("C", self.r.a & 0x40)
        self.r.setFlag("V", bool(self.r.a & 0x40) ^ bool(self.r.a & 0x20))
        self.r.ZN(self.r.a)
        # self.AND(v)
        # self.ROR("a")

    def ASR(self, v) -> None:  # ALR
        """
        AND oper + LSR

        .. important:: Illegal Opcode

        .. note:: Also known as: `ALR`

        +----------------------+---------------+
        | Function n n n n n   | `N Z C I D V` |
        +======================+===============+
        | A AND oper           | `N Z C - - -` |
        +----------------------+               |
        | 0 -> [76543210] -> C |               |
        +----------------------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = (((self.r.a & v) & 0xFF) >> 1)
        self.r.setFlag("C", self.r.a & 0x0001)
        self.r.ZN(self.r.a)
        # self.AND(v)
        # self.LSR("a")

    def ATX(self, v: int) -> None:  # LXA, OAL
        """
        Store * AND oper in A and X

        .. important:: Illegal Opcode

        .. note:: Also known as: `LXA`, `OAL`

        .. DANGER:: Highly unstable
           Involves a "magic" constant

        .. seealso::
           :py:meth:`.XAA`

        +------------------+---------------+
        | Function         | `N Z C I D V` |
        +==================+===============+
        | (A OR CONST) AND | `N Z - - - -` |
        | oper -> A -> X   |               |
        +------------------+---------------+

        :param int v: 8 bit value
        """
        # self.AND(v)
        # self.T(("a", "x"))
        value = (self.r.a ^ self.magic) & v
        self.r.a = value
        self.r.x = value
        self.r.ZN(value)

    def AXA(self, a: int) -> None:  # SHA
        """
        Stores A AND X AND (high-byte of addr. + 1) at addr.

        .. important:: Illegal Opcode

        .. note:: Also known as: `SHA`

        .. Warning:: Unstable
           Sometimes `AND (H+1)` is dropped, page boundary crossings may not
           work (with the high-byte of the value used as the high-byte of the
           address)

        There are a few illegal opcodes which and the high
        bit of the address with registers and write the values
        back into that address.  These operations are
        particularly screwy.  These posts are used as reference
        but I am unsure whether they are correct.

        - http://forums.nesdev.com/viewtopic.php?f=3&t=3831&start=30#p113343

        - http://forums.nesdev.com/viewtopic.php?f=3&t=10698

        +------------------+---------------+
        | Function         | `N Z C I D V` |
        +==================+===============+
        | A AND X AND      | `- - - - - -` |
        | (H+1) -> M       |               |
        +------------------+---------------+

        :param int a: 16 bit address
        """
        o = (a - self.r.y) & 0xFFFF
        low = o & 0xFF
        high = o >> 8

        self.increment_cycle_count()
        if (low + self.r.y) > 0xFF:  # crossed page
            a = ((high & self.r.x) << 8) + low + self.r.y
        else:
            a = (high << 8) + low + self.r.y

        v = self.r.a & self.r.x & (high + 1)
        self.writeByte(a, v)

    def AXS(self, v: int) -> None:  # SBX, SAX
        """
        CMP and DEX at once, sets flags like CMP

        .. important:: Illegal Opcode

        .. note:: Also known as: `SBX`, `SAX`

        +-----------------------+---------------+
        | Function              | `N Z C I D V` |
        +=======================+===============+
        | (A AND X) - oper -> X | `- - - - - -` |
        +-----------------------+---------------+

        :param int v: 8 bit value
        """
        o = self.r.a & self.r.x
        self.r.x = (o - v) & 0xFF

        self.r.setFlag("C", v <= o)
        self.r.ZN(self.r.x)

    def DCP(self, a: int) -> None:  # DCM
        """
        DEC oper + CMP oper

        .. important:: Illegal Opcode

        .. note:: Also known as: `DCN`

        +------------+---------------+
        | Function   | `N Z C I D V` |
        +============+===============+
        | M - 1 -> M | `N Z C - - -` |
        +------------+               |
        | A - M      |               |
        +------------+---------------+

        :param int a: 16 bit address
        """
        self.DEC(a)
        self.CMP(self.mmu.read(a))

    def ISC(self, a: int) -> None:  # ISB, INS
        """
        INC oper + SBC oper

        .. important:: Illegal Opcode

        .. note:: Also known as: `ISB`, `INS`

        +----------------------+---------------+
        | Function             | `N Z C I D V` |
        +======================+===============+
        | M + 1 -> M           | `N Z C - - V` |
        +----------------------+               |
        | A - M - (C - 1) -> A |               |
        +----------------------+---------------+

        :param int a: 16 bit address
        """
        self.INC(a)
        self.SBC(self.mmu.read(a))

    def KIL(self, _) -> None:  # JAM, HLT
        """
        Freezes the CPU

        .. important:: Illegal Opcode

        .. note:: Also known as: `JAM`, `HLT`

        .. Warning:: This instruction freezes the CPU.
           The processor will be trapped infinitely in T1 phase with $FF on
           the data bus. **Reset required**

        :param int _: Ignored
        """
        self.running = False

    def LAR(self, v: int) -> None:  # LAE, LAS
        """
        LDA/TSX oper

        .. important:: Illegal Opcode

        .. note:: Also known as: `LAE`, `LAS`

        +----------------------+---------------+
        | Function             | `N Z C I D V` |
        +======================+===============+
        | M AND SP -> A, X, SP | `N Z - - - -` |
        +----------------------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = self.r.x = self.r.s = self.r.s & v
        self.r.ZN(self.r.a)

    def LAX(self, v: int) -> None:
        """
        LDA oper + LDX oper

        .. important:: Illegal Opcode

        +-------------+---------------+
        | Function    | `N Z C I D V` |
        +=============+===============+
        | M -> A -> X | `N Z - - - -` |
        +-------------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = v & 0xFF
        self.r.x = v & 0xFF
        self.r.ZN(self.r.a)

    def RLA(self, a: int) -> None:
        """
        ROL oper + AND oper

        .. important:: Illegal Opcode

        +--------------------------+---------------+
        | Function                 | `N Z C I D V` |
        +==========================+===============+
        | M = C <- [76543210] <- C | `N Z C - - -` |
        +--------------------------+               |
        | A AND M -> A             |               |
        +--------------------------+---------------+

        :param int a: 16 bit address
        """
        self.ROL(a)
        self.AND(self.mmu.read(a))

    def RRA(self, a: int) -> None:
        """
        ROL oper + ADC oper

        .. important:: Illegal Opcode

        +--------------------------+---------------+
        | Function                 | `N Z C I D V` |
        +==========================+===============+
        | M = C -> [76543210] -> C | `N Z C - - V` |
        +--------------------------+               |
        | A + M + C -> A, C        |               |
        +--------------------------+---------------+

        :param int a: 16 bit address
        """

        self.ROR(a)
        self.ADC(self.mmu.read(a))

    def SLO(self, a: int) -> None:  # ASO
        """
        ASL oper + ORA oper

        .. important:: Illegal Opcode

        .. note:: Also known as: `ASO`

        +--------------------------+---------------+
        | Function                 | `N Z C I D V` |
        +==========================+===============+
        | M = C <- [76543210] <- 0 | `N Z C - - -` |
        +--------------------------+               |
        | A OR M -> A              |               |
        +--------------------------+---------------+

        :param int a: 16 bit address
        """
        self.ASL(a)
        self.ORA(self.mmu.read(a))

    def SRE(self, a: int) -> None:  # LSE
        """
        LSR oper + EOR oper

        .. important:: Illegal Opcode

        .. note:: Also known as: `LSE`

        +--------------------------+---------------+
        | Function                 | `N Z C I D V` |
        +==========================+===============+
        | M = 0 -> [76543210] -> 0 | `N Z C - - -` |
        +--------------------------+               |
        | A EOR M -> A             |               |
        +--------------------------+---------------+

        :param int a: 16 bit address
        """
        self.LSR(a)
        self.EOR(self.mmu.read(a))

    def SXA(self, a: int) -> None:  # SHX, XAS
        """
        Stores X AND (high-byte of addr. + 1) at addr.

        .. important:: Illegal Opcode

        .. note:: Also known as: `SHX`, `XAS`

        +------------------+---------------+
        | Function         | `N Z C I D V` |
        +==================+===============+
        | X AND (H+1) -> M | `- - - - - -` |
        +------------------+---------------+

        .. seealso::
           :py:meth:`.AXA`

        :param int a: 16 bit address
        """
        o = (a - self.r.y) & 0xFFFF
        low = o & 0xFF
        high = o >> 8
        self.increment_cycle_count()
        if low + self.r.y > 0xFF:  # crossed page
            a = ((high & self.r.x) << 8) + low + self.r.y
        else:
            a = (high << 8) + low + self.r.y

        v = self.r.x & (high + 1)
        self.writeByte(a, v)

    def SYA(self, a: int) -> None:  # SHY, SAY
        """
        Stores Y AND (high-byte of addr. + 1) at addr.

        .. important:: Illegal Opcode

        .. note:: Also known as: `SHY`, `SAY`

        +------------------+---------------+
        | Function         | `N Z C I D V` |
        +==================+===============+
        | Y AND (H+1) -> M | `- - - - - -` |
        +------------------+---------------+

        .. seealso::
           :py:meth:`.AXA`

        :param int a: 16 bit address
        """
        o = (a - self.r.x) & 0xFFFF
        low = o & 0xFF
        high = o >> 8
        self.increment_cycle_count()
        if low + self.r.x > 0xFF:  # crossed page
            a = ((high & self.r.y) << 8) + low + self.r.x
        else:
            a = (high << 8) + low + self.r.x

        v = self.r.y & (high + 1)
        self.writeByte(a, v)

    def XAA(self, v: int) -> None:  # ANE
        """
        \\* AND X + AND oper

        .. important:: Illegal Opcode

        .. note:: Also known as: `ANE`

        | Another very wonky operation.
        | It's fully described here: `6502_Opcode_8B_XAA_ANE`_

        "magic" varies by version of the processor.
        `0xEE` seems to be common.

        The formula is: `A = (A | magic) & X & imm`

        .. DANGER:: Highly unstable
           Involves a "magic" constant

           A base value in A is determined based on the
           contets of A and a constant, which may be
           typically `$00`, `$FF`, `$EE`, etc.

           The value of this constant depends on temerature,
           the chip series, and maybe other factors, as well.

           In order to eliminate these uncertaincies from
           the equation, use either 0 as the operand or a
           value of `$FF` in the accumulator.

        .. _6502_Opcode_8B_XAA_ANE:
            http://visual6502.org/wiki/index.php?title=6502_Opcode_8B_%28XAA,_ANE%29  # noqa E501

        +--------------------+---------------+
        | Function           | `N Z C I D V` |
        +====================+===============+
        | (A OR CONST) AND X | `N Z - - - -` |
        | AND oper -> A      |               |
        +--------------------+---------------+

        :param int v: 8 bit value
        """
        self.r.a = (self.r.a | self.magic) & self.r.x & v
        self.r.ZN(self.r.a)

    def XAS(self, a: int) -> None:  # SHS, TAS
        """
        Puts A AND X in SP and stores A AND X AND (high-byte of addr. + 1)

        .. important:: Illegal Opcode

        .. note:: Also known as: `SHS`, `TAS`

        +--------------------------+---------------+
        | Function                 | `N Z C I D V` |
        +==========================+===============+
        | A AND X -> SP            | `- - - - - -` |
        +--------------------------+               |
        | A AND X AND (H + 1) -> M |               |
        +--------------------------+---------------+

        :param int a: 16 bit address
        """
        # First set the stack pointer's value
        self.r.s = self.r.a & self.r.x

        # Then write to memory using the new value of the stack pointer
        o = (a - self.r.y) & 0xFFFF
        low = o & 0xFF
        high = o >> 8
        self.increment_cycle_count()
        if low + self.r.y > 0xFF:  # crossed page
            a = ((high & self.r.s) << 8) + low + self.r.y
        else:
            a = (high << 8) + low + self.r.y

        v = self.r.s & (high + 1)
        self.writeByte(a, v)

    # endregion Illegal/Undocumented Operations



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_suites
----------------------------------

Tests for `py65emu` module.
"""


import os
import unittest

from py65emu.cpu import CPU, Registers
from py65emu.mmu import MMU
from py65emu.debug import Debug


class NesTestError(Exception):
    def __init__(
        self,
        message: str,
        actual: CPU,
        expected: Registers | None = None,
        *args: object
    ) -> None:
        super().__init__(*[message, actual, expected] + [*args])
        self.message = message
        self.actual = actual
        self.expected = expected

    def __str__(self) -> str:
        d = Debug(self.actual)
        d.memdump(0x0647)
        return (
            f"C: {self.actual.cc_total:d} "
            f"OP: ${self.actual.r.pc:0>4x}: {self.message}"
            f"\nActual:   {self.actual.r!r}\nExpected: {self.expected!r}"
        )


class NesTestRom(unittest.TestCase):
    def setUp(self):
        self.c = self.load_cpu()
        # self.c.reset()
        # self.c.r.pc = 0xC000
        self.reg = self.load_nestest_log()

    def tearDown(self):
        # print("\n\nEXITING:", f"Cycles: {self.c.cc_total:d}\n\n")
        # Debug.crash_dump(self.c)
        pass

    def load_cpu(self) -> "CPU":
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            # "nestest_mod2.nes"
            "nestest.nes"
        )

        with open(path, "rb") as f:
            mmu = MMU(
                [
                    (0x0000, 0x800),  # RAM
                    (0x2000, 0x8),  # PPU
                    (0x4000, 0x18),
                    (0x8000, 0xC000, True, f, 0x3FF0),  # ROM
                ]
            )

        c = CPU(mmu=mmu, pc=0xC000, disable_bcd=True)
        c.r.s = 0xFD
        return c

    def load_nestest_log(self) -> dict[int, "Registers"]:
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "nestest.log"
        )
        registers = {}
        with open(path, "r") as file:
            for f in file.readlines():
                cycle = int(f[90:])
                r = Registers(int(f[:4], 16))
                r.a = int(f[50:52], 16)
                r.x = int(f[55:57], 16)
                r.y = int(f[60:62], 16)
                r.p = int(f[65:67], 16)
                r.s = int(f[71:73], 16)

                registers[cycle] = r

        return registers

    def test_nestest(self):
        # cycle = (self.c.cc_total - self.c.cc)
        cycle = self.c.cc_total
        self.checkCycle(cycle)

        while self.c.r.pc != 0xC66E:
            self.c.step()

            cycle = self.c.cc_total
            self.checkCycle(cycle)

            # Hmm, a diff in 41 cycles...
            """
            self.assertLessEqual(
                self.c.cc_total, 26554, "Too many cycles!"
            )
            """
            self.assertLessEqual(
                self.c.cc_total, 30000, "Too many cycles!"
            )

    def checkCycle(self, cycle: int) -> None:
        pc = "OP: {:0>4x}".format(self.c.r.pc)

        legal_op_error = self.c.mmu.read(0x0002)
        illegal_op_error = self.c.mmu.read(0x0003)

        self.assertEqual(
            legal_op_error,
            0x00,
            f"Cycle {cycle:d} - Caught Error in index 0x02: "
            f"0x{legal_op_error:0>2X} - PC 0x{pc:s}"
        )
        self.assertEqual(
            illegal_op_error,
            0x00,
            f"Cycle {cycle:d} - Caught Error in index 0x03: "
            f"0x{illegal_op_error:0>2X} - PC 0x{pc:s}"
        )

        if cycle not in self.reg:
            raise NesTestError(
                message="Cycle not found in nestest.log",
                actual=self.c
            )

        reg = self.reg[cycle]

        # Accumulator
        if self.c.r.a != reg.a:
            raise NesTestError(
                message="Error on Accumulator", actual=self.c, expected=reg
            )

        # X Register
        if self.c.r.x != reg.x:
            raise NesTestError(
                message="Error on X Register", actual=self.c, expected=reg
            )

        # Y Register
        if self.c.r.y != reg.y:
            raise NesTestError(
                message="Error on Y Register", actual=self.c, expected=reg
            )

        # P Register
        if self.c.r.p != reg.p:
            raise NesTestError(
                message="Error on P Register", actual=self.c, expected=reg
            )

        # Stack Pointer
        if self.c.r.s != reg.s:
            raise NesTestError(
                message="Error on Stack Pointer", actual=self.c, expected=reg
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_cpu
----------------------------------

Tests for `py65emu` module.
"""

import os
import unittest

from py65emu.cpu import CPU, FlagBit
from py65emu.mmu import MMU


class TestCPU(unittest.TestCase):
    def _cpu(
        self,
        ram=(0, 0x200, False),
        rom=(0x1000, 0x100),
        romInit=None,
        pc=0x1000,
    ):
        return CPU(MMU([ram, rom + (True, romInit)]), pc)

    def setUp(self):
        pass

    def test_fromBCD(self):
        c = self._cpu()
        self.assertEqual(c.fromBCD(0), 0)
        self.assertEqual(c.fromBCD(0x05), 5)
        self.assertEqual(c.fromBCD(0x11), 11)
        self.assertEqual(c.fromBCD(0x99), 99)

    def test_toBCD(self):
        c = self._cpu()
        self.assertEqual(c.toBCD(0), 0)
        self.assertEqual(c.toBCD(5), 0x05)
        self.assertEqual(c.toBCD(11), 0x11)
        self.assertEqual(c.toBCD(99), 0x99)

    def test_fromTwosCom(self):
        c = self._cpu()
        self.assertEqual(c.fromTwosCom(0x00), 0)
        self.assertEqual(c.fromTwosCom(0x01), 1)
        self.assertEqual(c.fromTwosCom(0x7F), 127)
        self.assertEqual(c.fromTwosCom(0xFF), -1)
        self.assertEqual(c.fromTwosCom(0x80), -128)

    def test_nextByte(self):
        c = self._cpu(romInit=[1, 2, 3])
        self.assertEqual(c.nextByte(), 1)
        self.assertEqual(c.nextByte(), 2)
        self.assertEqual(c.nextByte(), 3)
        self.assertEqual(c.nextByte(), 0)

    def test_nextWord(self):
        c = self._cpu(romInit=[1, 2, 3, 4, 5, 9, 10])
        self.assertEqual(c.nextWord(), 0x0201)
        c.nextByte()
        self.assertEqual(c.nextWord(), 0x0504)
        self.assertEqual(c.nextWord(), 0x0A09)

    def test_zeropage_addressing(self):
        c = self._cpu(romInit=[1, 2, 3, 4, 5])
        self.assertEqual(c.z_a(), 1)
        c.r.x = 0
        self.assertEqual(c.zx_a(), 2)
        c.r.x = 1
        self.assertEqual(c.zx_a(), 4)
        c.r.y = 0
        self.assertEqual(c.zy_a(), 4)
        c.r.y = 1
        self.assertEqual(c.zy_a(), 6)

    def test_absolute_addressing(self):
        c = self._cpu(romInit=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertEqual(c.a_a(), 0x0201)

        c.r.x = 0
        c.cc = 0
        self.assertEqual(c.ax_a(), 0x0403)
        self.assertEqual(c.cc, 2)
        c.r.x = 0xFF
        c.cc = 0
        self.assertEqual(c.ax_a(), 0x0605 + 0xFF)
        self.assertEqual(c.cc, 2)

        c.r.y = 0
        c.cc = 0
        self.assertEqual(c.ay_a(), 0x0807)
        self.assertEqual(c.cc, 2)
        c.r.y = 0xFF
        c.cc = 0
        self.assertEqual(c.ay_a(), 0x0A09 + 0xFF)
        self.assertEqual(c.cc, 2)

    def test_indirect_addressing(self):
        c = self._cpu(
            romInit=[
                0x06,
                0x10,
                0xFF,
                0x10,
                0x00,
                0x00,
                0xF0,
                0x00,
            ]
        )

        self.assertEqual(c.i_a(), 0x00F0)
        self.assertEqual(c.i_a(), 0x0600)

        c.r.y = 0x05
        c.mmu.write(0x00, 0x21)
        c.mmu.write(0x01, 0x43)
        self.assertEqual(c.iy_a(), 0x4326)

        c.r.x = 0x02
        c.mmu.write(0x02, 0x34)
        c.mmu.write(0x03, 0x12)
        self.assertEqual(c.ix_a(), 0x1234)

    def test_stack(self):
        c = self._cpu()
        c.stackPush(0x10)
        self.assertEqual(c.stackPop(), 0x10)
        c.stackPushWord(0x0510)
        self.assertEqual(c.stackPopWord(), 0x0510)
        self.assertEqual(c.stackPop(), 0x00)
        c.stackPush(0x00)
        c.stackPushWord(0x0510)
        self.assertEqual(c.stackPop(), 0x10)
        self.assertEqual(c.stackPop(), 0x05)

    def test_adc(self):
        c = self._cpu(romInit=[1, 2, 250, 3, 100, 100])
        # immediate
        c.execute([0x69])
        self.assertEqual(c.r.a, 1)
        c.execute([0x69])
        self.assertEqual(c.r.a, 3)
        c.execute([0x69])
        self.assertEqual(c.r.a, 253)
        self.assertTrue(c.r.getFlag(FlagBit.N))
        c.r.clearFlags()
        c.execute([0x69])
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.Z))
        c.r.clearFlags()
        c.execute([0x69, 0x69])
        self.assertTrue(c.r.getFlag(FlagBit.V))

    def test_adc_decimal(self):
        c = self._cpu(romInit=[0x01, 0x55, 0x50])
        c.r.setFlag(FlagBit.D)

        c.execute([0x69])
        self.assertEqual(c.r.a, 0x01)
        c.execute([0x69])
        self.assertEqual(c.r.a, 0x56)
        c.execute([0x69])
        self.assertEqual(c.r.a, 0x06)
        self.assertTrue(c.r.getFlag(FlagBit.C))

    def test_and(self):
        c = self._cpu(romInit=[0xFF, 0xFF, 0x01, 0x2])

        c.r.a = 0x00
        c.execute([0x29])
        self.assertEqual(c.r.a, 0)

        c.r.a = 0xFF
        c.execute([0x29])
        self.assertEqual(c.r.a, 0xFF)

        c.r.a = 0x01
        c.execute([0x29])
        self.assertEqual(c.r.a, 0x01)

        c.r.a = 0x01
        c.execute([0x29])
        self.assertEqual(c.r.a, 0x00)

    def test_asl(self):
        c = self._cpu(romInit=[0x00])

        c.r.a = 1
        c.execute([0x0A])
        self.assertEqual(c.r.a, 2)

        c.mmu.write(0, 4)
        c.execute([0x06])
        self.assertEqual(c.mmu.read(0), 8)

    def test_bit(self):
        c = self._cpu(romInit=[0x00, 0x00, 0x10])
        c.mmu.write(0, 0xFF)
        c.r.a = 1

        c.execute([0x24])  # Zero page
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.N))
        self.assertTrue(c.r.getFlag(FlagBit.V))

        c.execute([0x2C])  # Absolute
        self.assertTrue(c.r.getFlag(FlagBit.Z))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.V))

    def test_brk(self):
        c = self._cpu()
        c.mmu.addBlock(0xFFFE, 0x2, True, [0x34, 0x12])
        c.r.p = 0b11101111
        c.execute([0x00])
        # self.assertTrue(c.r.getFlag(FlagBit.B))
        self.assertTrue(c.r.getFlag(FlagBit.I))
        self.assertEqual(c.r.pc, 0x1234)
        self.assertEqual(c.stackPop(), 255)
        self.assertEqual(c.stackPopWord(), 0x1001)

    def test_branching(self):
        c = self._cpu(romInit=[0x01, 0x00, 0x00, 0xFC])
        c.execute([0x10])
        self.assertEqual(c.r.pc, 0x1002)
        c.execute([0x70])
        self.assertEqual(c.r.pc, 0x1003)
        c.r.setFlag(FlagBit.C)
        c.execute([0xB0])
        self.assertEqual(c.r.pc, 0x1000)
        c.execute([0xD0])
        self.assertEqual(c.r.pc, 0x1002)

    def test_cmp(self):
        c = self._cpu(romInit=[0x0F, 0x10, 0x11, 0xFE, 0xFF, 0x00, 0x7F])

        c.r.a = 0x10
        c.execute([0xC9])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xC9])
        self.assertTrue(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xC9])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertFalse(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.N))

        c.r.a = 0xFF
        c.execute([0xC9])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xC9])
        self.assertTrue(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xC9])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.N))
        c.execute([0xC9])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.N))

    def test_cpx(self):
        c = self._cpu(romInit=[0x0F, 0x10, 0x11])

        c.r.x = 0x10
        c.execute([0xE0])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xE0])
        self.assertTrue(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xE0])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertFalse(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.N))

    def test_cpy(self):
        c = self._cpu(romInit=[0x0F, 0x10, 0x11])

        c.r.y = 0x10
        c.execute([0xC0])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xC0])
        self.assertTrue(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        c.execute([0xC0])
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertFalse(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.N))

    def test_dec(self):
        c = self._cpu(romInit=[0x00])
        c.execute([0xC6])
        self.assertEqual(c.mmu.read(0x00), 0xFF)

    def test_dex(self):
        c = self._cpu()
        c.execute([0xCA])
        self.assertEqual(c.r.x, 0xFF)

    def test_dey(self):
        c = self._cpu()
        c.execute([0x88])
        self.assertEqual(c.r.y, 0xFF)

    def test_eor(self):
        c = self._cpu(romInit=[0x0F, 0xF0, 0xFF])

        c.execute([0x49])
        self.assertEqual(c.r.a, 0x0F)
        c.execute([0x49])
        self.assertEqual(c.r.a, 0xFF)
        c.execute([0x49])
        self.assertEqual(c.r.a, 0x00)

    def test_flag_ops(self):
        c = self._cpu()

        c.execute([0x38])
        self.assertTrue(c.r.getFlag(FlagBit.C))
        c.execute([0x78])
        self.assertTrue(c.r.getFlag(FlagBit.I))
        c.execute([0xF8])
        self.assertTrue(c.r.getFlag(FlagBit.D))

        c.r.setFlag(FlagBit.V)

        c.execute([0x18])
        self.assertFalse(c.r.getFlag(FlagBit.C))
        c.execute([0x58])
        self.assertFalse(c.r.getFlag(FlagBit.I))
        c.execute([0xB8])
        self.assertFalse(c.r.getFlag(FlagBit.V))
        c.execute([0xD8])
        self.assertFalse(c.r.getFlag(FlagBit.D))

    def test_inc(self):
        c = self._cpu(romInit=[0x00])
        c.execute([0xE6])
        self.assertEqual(c.mmu.read(0x00), 0x01)

    def test_inx(self):
        c = self._cpu()
        c.execute([0xE8])
        self.assertEqual(c.r.x, 0x01)

    def test_iny(self):
        c = self._cpu()
        c.execute([0xC8])
        self.assertEqual(c.r.y, 0x01)

    def test_jmp(self):
        c = self._cpu(romInit=[0x00, 0x10])

        c.execute([0x4C])
        self.assertEqual(c.r.pc, 0x1000)

        c.execute([0x6C])
        self.assertEqual(c.r.pc, 0x1000)

    def test_jsr(self):
        c = self._cpu(romInit=[0x00, 0x10])

        c.execute([0x20])
        self.assertEqual(c.r.pc, 0x1000)
        self.assertEqual(c.stackPopWord(), 0x1001)

    def test_lda(self):
        c = self._cpu(romInit=[0x01])
        c.execute([0xA9])
        self.assertEqual(c.r.a, 0x01)

    def test_ldx(self):
        c = self._cpu(romInit=[0x01])
        c.execute([0xA2])
        self.assertEqual(c.r.x, 0x01)

    def test_ldy(self):
        c = self._cpu(romInit=[0x01])
        c.execute([0xA0])
        self.assertEqual(c.r.y, 0x01)

    def test_lsr(self):
        c = self._cpu(romInit=[0x00])

        c.r.a = 0x02

        c.execute([0x4A])
        self.assertEqual(c.r.a, 0x01)
        self.assertFalse(c.r.getFlag(FlagBit.C))

        c.execute([0x4A])
        self.assertEqual(c.r.a, 0x00)
        self.assertTrue(c.r.getFlag(FlagBit.C))

        c.mmu.write(0x00, 0x02)
        c.execute([0x46])
        self.assertEqual(c.mmu.read(0x00), 0x01)

    def test_ora(self):
        c = self._cpu(romInit=[0x0F, 0xF0, 0xFF])
        c.execute([0x09])
        self.assertEqual(c.r.a, 0x0F)
        c.execute([0x09])
        self.assertEqual(c.r.a, 0xFF)
        c.execute([0x09])
        self.assertEqual(c.r.a, 0xFF)

    def test_p(self):
        c = self._cpu()

        c.r.a = 0xCC
        c.execute([0x48])
        self.assertEqual(c.stackPop(), 0xCC)

        c.r.p = 0xFF
        c.execute([0x08])
        self.assertEqual(c.stackPop(), 0xFF)

        c.r.a = 0x00
        c.stackPush(0xDD)
        c.execute([0x68])
        self.assertEqual(c.r.a, 0xDD)

        c.r.p = 0x20
        c.stackPush(0xFD)
        c.execute([0x28])
        self.assertEqual(c.r.p, 0xED)

    def test_rol(self):
        c = self._cpu(romInit=[0x00])

        c.r.a = 0xFF
        c.execute([0x2A])
        self.assertEqual(c.r.a, 0xFE)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        c.execute([0x2A])
        self.assertEqual(c.r.a, 0xFD)
        self.assertTrue(c.r.getFlag(FlagBit.C))

        c.execute([0x26])
        self.assertEqual(c.mmu.read(0x00), 0x01)
        self.assertFalse(c.r.getFlag(FlagBit.C))

    def test_ror(self):
        c = self._cpu(romInit=[0x00])

        c.r.a = 0xFF
        c.execute([0x6A])
        self.assertEqual(c.r.a, 0x7F)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        c.execute([0x6A])
        self.assertEqual(c.r.a, 0xBF)
        self.assertTrue(c.r.getFlag(FlagBit.C))

        c.execute([0x66])
        self.assertEqual(c.mmu.read(0x00), 0x80)
        self.assertFalse(c.r.getFlag(FlagBit.C))

    def test_rti(self):
        c = self._cpu()

        c.stackPushWord(0x1234)
        c.stackPush(0xFD)

        c.execute([0x40])
        self.assertEqual(c.r.pc, 0x1234)
        self.assertTrue(c.r.getFlag(FlagBit.N))
        self.assertTrue(c.r.getFlag(FlagBit.V))
        self.assertTrue(c.r.getFlag(FlagBit.B))
        self.assertTrue(c.r.getFlag(FlagBit.D))
        self.assertTrue(c.r.getFlag(FlagBit.I))
        self.assertFalse(c.r.getFlag(FlagBit.Z))
        self.assertTrue(c.r.getFlag(FlagBit.C))

    def test_rts(self):
        c = self._cpu()
        c.stackPushWord(0x1234)
        c.execute([0x60])
        self.assertEqual(c.r.pc, 0x1235)

    def test_sbc(self):
        c = self._cpu(romInit=[0x10, 0x01, 0x51, 0x80, 0x12, 0x13, 0x02, 0x21])

        c.r.a = 0x15
        c.r.setFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0x05)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.V))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

        c.r.a = 0xFF
        c.r.setFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0xFE)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.V))
        self.assertTrue(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

        c.r.a = 0x50
        c.r.setFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0xFF)
        self.assertFalse(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.V))
        self.assertTrue(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

        c.r.a = 0x01
        c.r.setFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0x81)
        self.assertFalse(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.V))
        self.assertTrue(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

        # decimal mode test
        c.r.setFlag(FlagBit.D)

        c.r.a = 0x46
        c.r.setFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0x34)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.V))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

        c.r.a = 0x40
        c.r.setFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0x27)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.V))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

        c.r.a = 0x32
        c.r.clearFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0x29)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.V))
        self.assertFalse(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

        c.r.a = 0x12
        c.r.setFlag(FlagBit.C)
        c.execute([0xE9])
        self.assertEqual(c.r.a, 0x91)
        self.assertFalse(c.r.getFlag(FlagBit.C))
        self.assertFalse(c.r.getFlag(FlagBit.V))
        self.assertTrue(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.Z))

    def test_sta(self):
        c = self._cpu(romInit=[0x00])
        c.r.a = 0xF0
        c.execute([0x85])
        self.assertEqual(c.mmu.read(0x00), 0xF0)

    def test_stx(self):
        c = self._cpu(romInit=[0x00])
        c.r.x = 0xF0
        c.execute([0x86])
        self.assertEqual(c.mmu.read(0x00), 0xF0)

    def test_sty(self):
        c = self._cpu(romInit=[0x00])
        c.r.y = 0xF0
        c.execute([0x84])
        self.assertEqual(c.mmu.read(0x00), 0xF0)

    def test_t(self):
        c = self._cpu()

        c.r.a = 0xF0
        c.execute([0xAA])
        self.assertEqual(c.r.x, 0xF0)

        c.r.x = 0x0F
        c.execute([0x8A])
        self.assertEqual(c.r.a, 0x0F)

        c.r.a = 0xFF
        c.execute([0xA8])
        self.assertEqual(c.r.y, 0xFF)

        c.r.y = 0x00
        c.execute([0x98])
        self.assertEqual(c.r.a, 0x00)

        c.r.x = 0xFF
        c.execute([0x9A])
        self.assertEqual(c.r.s, 0xFF)

        c.r.s = 0xF0
        c.execute([0xBA])
        self.assertEqual(c.r.x, 0xF0)

    def test_aac(self):
        c = self._cpu(romInit=[0xFF, 0xFF, 0x01, 0x2])

        c.r.a = 0x00
        c.execute([0x0B])
        self.assertEqual(c.r.a, 0)
        self.assertFalse(c.r.getFlag(FlagBit.N))
        self.assertFalse(c.r.getFlag(FlagBit.C))

        c.r.a = 0xFF
        c.execute([0x2B])
        self.assertEqual(c.r.a, 0xFF)
        self.assertTrue(c.r.getFlag(FlagBit.N))
        self.assertTrue(c.r.getFlag(FlagBit.C))

    def test_aax(self):
        c = self._cpu(romInit=[0x00, 0x00])

        c.r.a = 0xF0
        c.r.x = 0xF0
        c.execute([0x87])
        self.assertEqual(c.mmu.read(0x00), 0xF0)

    def test_arr(self):
        c = self._cpu(romInit=[0x80])

        c.r.a = 0xFF
        c.r.setFlag(FlagBit.C)
        c.execute([0x6B])
        self.assertEqual(c.r.a, 0xC0)
        self.assertTrue(c.r.getFlag(FlagBit.C))
        self.assertTrue(c.r.getFlag(FlagBit.V))

    def test_asr(self):
        c = self._cpu(romInit=[0x80])

        c.r.a = 0xFF
        c.r.setFlag(FlagBit.C)
        c.execute([0x4B])
        self.assertEqual(c.r.a, 0x40)

    def test_atx(self):
        c = self._cpu(romInit=[0xF8])

        c.r.a = 0x1F
        c.execute([0xAB])
        # self.assertEqual(c.r.x, 0x18)
        self.assertEqual(c.r.a, 0xf0)
        self.assertEqual(c.r.x, 0xf0)

    def test_axa(self):
        c = self._cpu(ram=(0, 0x400, False), romInit=[0xFF, 0x01])

        c.r.a = c.r.x = 0xFF
        c.r.y = 0x01
        c.execute([0x9F])

        self.assertEqual(c.mmu.read(0x200), 0x02)

    def test_axs(self):
        c = self._cpu(romInit=[0x02])

        c.r.a = 0xF0
        c.r.x = 0x0F
        c.execute([0xCB])
        self.assertEqual(c.r.x, 0xFE)

    def test_dcp(self):
        c = self._cpu(romInit=[0x01])
        c.r.a = 0xFF
        c.execute([0xC7])
        self.assertEqual(c.mmu.read(0x01), 0xFF)
        self.assertTrue(c.r.getFlag(FlagBit.Z))

    def test_isc(self):
        c = self._cpu(romInit=[0x01])
        c.r.a = 0xFF
        c.r.setFlag(FlagBit.C)
        c.execute([0xE7])
        self.assertEqual(c.mmu.read(0x01), 0x01)
        self.assertEqual(c.r.a, 0xFE)

    def test_kil(self):
        c = self._cpu()
        c.execute([0x02])
        self.assertFalse(c.running)

    def test_lar(self):
        c = self._cpu(romInit=[0x01, 0x00])
        c.r.y = 0x01
        c.mmu.write(0x02, 0xF0)

        c.execute([0xBB])
        self.assertEqual(c.r.a, 0xF0)
        self.assertEqual(c.r.x, 0xF0)
        self.assertEqual(c.r.s, 0xF0)

    def test_lax(self):
        c = self._cpu(romInit=[0x01])
        c.mmu.write(0x01, 0xF0)
        c.execute([0xA7])
        self.assertEqual(c.r.a, 0xF0)
        self.assertEqual(c.r.x, 0xF0)

    def test_rla(self):
        c = self._cpu(romInit=[0x01])
        c.mmu.write(0x01, 0x01)
        c.r.a = 0x06
        c.r.setFlag(FlagBit.C)
        c.execute([0x27])
        self.assertEqual(c.mmu.read(0x01), 0x03)
        self.assertEqual(c.r.a, 0x02)

    def test_rra(self):
        c = self._cpu(romInit=[0x01])
        c.mmu.write(0x01, 0x01)
        c.r.a = 0x06
        c.r.setFlag(FlagBit.C)
        c.execute([0x67])
        self.assertEqual(c.mmu.read(0x01), 0x80)
        self.assertEqual(c.r.a, 0x87)

    def test_rra2(self):
        c = self._cpu(romInit=[0x01])
        c.mmu.write(0x01, 0x02)
        c.r.a = 0x06
        c.r.setFlag(FlagBit.C)
        c.execute([0x47])
        self.assertEqual(c.mmu.read(0x01), 0x01)
        self.assertEqual(c.r.a, 0x07)

    def test_slo(self):
        c = self._cpu(romInit=[0x01])
        c.mmu.write(0x01, 0x01)
        c.r.a = 0x06
        c.r.setFlag(FlagBit.C)
        c.execute([0x07])
        self.assertEqual(c.mmu.read(0x01), 0x02)
        self.assertEqual(c.r.a, 0x06)

    def test_sxa(self):
        c = self._cpu(ram=(0, 0x400, False), romInit=[0xFF, 0x01])
        c.r.x = 0xFF
        c.r.y = 0x01
        c.execute([0x9E])

        self.assertEqual(c.mmu.read(0x200), 0x02)

    def test_sya(self):
        c = self._cpu(ram=(0, 0x400, False), romInit=[0xFF, 0x01])

        c.r.y = 0xFF
        c.r.x = 0x01
        c.execute([0x9C])

        self.assertEqual(c.mmu.read(0x200), 0x02)

    def test_xaa(self):
        c = self._cpu(romInit=[0xFF])

        c.r.a = 0b11111110
        c.r.x = 0b11101111

        c.execute([0x8B])

        self.assertEqual(c.r.a, 0b11101110)

    def test_xas(self):
        c = self._cpu(ram=(0, 0x400, False), romInit=[0xFF, 0x01])

        c.r.x = 0xFE
        c.r.a = 0x7F

        c.r.y = 0x01
        c.execute([0x9B])

        self.assertEqual(c.r.s, 0x7E)
        self.assertEqual(c.mmu.read(0x100), 0x02)

    def test_step(self):
        c = self._cpu(romInit=[0xA9, 0x55, 0x69, 0x22])
        c.step()
        self.assertEqual(c.r.a, 0x55)
        c.step()
        self.assertEqual(c.r.a, 0x77)

    def test_run_rom(self):
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "test_load_file.bin"
        )

        with open(path, "rb") as f:
            c = self._cpu(romInit=f)

        c.step()
        self.assertEqual(c.r.a, 0x55)

    def test_cycle_counting(self):
        # Adapted from @InvalidCo's test in #7
        c = self._cpu(
            romInit=[
                0xA0,
                0x80,
                0xB9,
                0x00,
                0x00,
                0xB9,
                0x80,
                0x00,
                0x18,
                0xB0,
                0xFF,
                0x90,
                0x01,
                0x90,
                0x8F,
            ]
        )

        expected_cycles = [2, 4, 5, 2, 2, 3, 4]

        for expected_cycle in expected_cycles:
            with self.subTest(expected_cycle=expected_cycle):
                c.step()
                self.assertEqual(c.cc, expected_cycle, f"{c.op.opcode:0>2x}")

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()



import array
import io
from typing import Sequence


class MemoryRangeError(ValueError):
    pass


class ReadOnlyError(TypeError):
    pass


class Block:
    def __init__(
        self,
        start: int,
        length: int,
        readonly: bool,
        default: int = 0
    ):
        """
        :param int start: The starting address for this block
        :param int length: The length of this block in bytes
        :param bool readOnly: Whether this block should be read only
                              (such as ROM) (default False)
        :param int default: Default value to initialize the block with
        """
        self.start = start
        self.length = length
        self.readonly = readonly
        self.default = default
        self._memory = array.array("B", [default] * length)

    def reset(self) -> None:
        if not self.readonly:
            self._memory = array.array("B", [self.default] * self.length)

    @property
    def end(self) -> int:
        return self.start + self.length

    def __setitem__(self, index: int, value: int) -> None:
        """
        Set value, absolute to the block, by the address in the block.

        :param int index: Address/Position to write to
        :param int value: Value to write
        :raises ReadOnlyError: If block is readonly
        :raises IndexError: If address is out of bounds for block
        """

        """
        if self.readonly:
            raise ReadOnlyError(
                "Memory section is readonly "
                "([${start:0>4x}] - [${length:0>4x}])"
                .format(start=0, length=self.length)
            )
        """

        if index < 0 or index >= self.length:
            raise IndexError(
                "Index [${index:0>4x}] is outside of memory range "
                "([${start:0>4x}] - [${length:0>4x}])".format(
                    index=index,
                    start=0x0,
                    length=self.length
                )
            )

        self._memory[index] = value

    def __getitem__(self, index: int) -> int:
        """
        Get value, relative to the block, by the address in the block.

        :param int index: Position to read from
        :raises IndexError: If address is out of bounds for block
        :rtype: int
        :return: Value of address (8 bit)
        """
        if index < 0 or index >= self.length:
            raise IndexError(
                "Index [${index:0>4x}] is outside of memory range "
                "([${start:0>4x}] - [${length:0>4x}])".format(
                    index=index,
                    start=0x0,
                    length=self.length
                )
            )

        return self._memory[index]

    def __repr__(self) -> str:
        result = ""
        for i in self._memory:
            result += "{:0>2x} ".format(i)

        return result

    def set(self, addr: int, value: int) -> None:
        """
        Set value, by the address in the block.

        :param int index: Address/Position to write to
        :param int value: Value to write
        :raises ReadOnlyError: If block is readonly
        :raises IndexError: If address is out of bounds for block
        """
        if self.readonly:
            raise ReadOnlyError(
                "Memory section is readonly "
                "(0x{s.start:0>4x} - 0x{s.end:0>4x})".format(s=self)
                )

        if addr < self.start or addr >= self.start + self.length:
            raise IndexError(
                "Address 0x{addr:0>4x} is outside of memory range "
                "(0x{s.start:0>4x} - 0x{s.end:0>4x})".format(
                    addr=addr, s=self
                )
            )

        self[addr - self.start] = value

    def get(self, addr: int) -> int:
        """
        Get value, by the address in the block.

        :param int addr: Address to read from
        :raises ReadOnlyError: If block is readonly
        :raises IndexError: If address is out of bounds for block
        :rtype: int
        :return: Value of address (8 bit)
        """
        if addr < self.start or addr >= self.start + self.length:
            raise IndexError(
                "Address 0x{addr:0>4x} is outside of memory range "
                "(0x{s.start:0>4x} - 0x{s.end:0>4x})".format(
                    addr=addr, s=self
                )
            )

        return self[addr - self.start]


class MMU:
    def __init__(self, blocks: Sequence[tuple] = []):
        """
        Initialize the MMU with the blocks specified in blocks.  blocks
        is a list of 5-tuples, (start, length, readonly, value, valueOffset).

        See `addBlock` for details about the parameters.
        """

        # Different blocks of memory stored seperately so that they can
        # have different properties.  Stored as dict of "start", "length",
        # "readonly" and "memory"
        self.blocks: list[Block] = []

        for b in blocks:
            if isinstance(b, tuple):
                self.addBlock(*b)

    def reset(self) -> None:
        """
        In all writeable blocks reset all values to zero.
        """
        for b in self.blocks:
            b.reset()

    def addBlock(
        self,
        start: int,
        length: int,
        readonly: bool = False,
        value: list | tuple | io.BufferedIOBase | None = None,
        valueOffset: int = 0
    ) -> None:
        """
        Add a block of memory to the list of blocks with the given start
        address and length; whether it is readonly or not; and the starting
        value as either a file pointer, binary value or list of unsigned
        integers.  If the block overlaps with an existing block an exception
        will be thrown.

        :param int start: The starting address of the block of memory
        :param int length: The length of the block in bytes
        :param bool readOnly: Whether the block should be read only
                              (such as ROM) (default False)
        :param value: The intial value for the block of memory. Used for
                      loading program data. (Default None)
        :param int valueOffset: Used when copying the above `value` into the
                                block to offset the location it is copied
                                into. For example, to copy byte 0 in `value`
                                into location 1000 in the block, set
                                valueOffest=1000. (Default 0)
        :type value: TextIO | list[int]
        """

        for b in self.blocks:
            if (
                start + length > b.start
                and start + length < b.start + b.length
            ) or (
                b.start + b.length > start
                and b.start + b.length < start + length
            ):
                raise MemoryRangeError()

        newBlock = Block(
            start=start,
            length=length,
            readonly=readonly
        )

        # raise TypeError(type(value))

        if (
            isinstance(value, list) or
            isinstance(value, tuple) or
            isinstance(value, str)
        ):
            for i in range(len(value)):
                newBlock[i + valueOffset] = value[i]
                # newBlock.set(i + valueOffset, value[i])
        elif isinstance(value, io.BufferedIOBase):
            a = array.array("B")
            a.frombytes(value.read())
            for i in range(len(a)):
                newBlock[i + valueOffset] = a[i]
                # newBlock.set(i + valueOffset, a[i])

        self.blocks.append(newBlock)

    def getBlock(self, addr: int) -> Block:
        """
        Get the block associated with the given address.

        :param int addr: Memory address to locate
        """

        for b in self.blocks:
            if addr >= b.start and addr < b.start + b.length:
                return b

        raise IndexError("Unable to locate position %{:0>4x}".format(addr))

    def write(self, addr: int, value: int) -> None:
        """
        Write a value to the given address if it is writeable.

        :param int index: Address/Position to write to
        :param int value: Value to write
        :raises ReadOnlyError: If block is readonly
        :raises IndexError: If address is out of bounds for block
        """
        b = self.getBlock(addr)
        b.set(addr, value & 0xFF)

    def read(self, addr: int) -> int:
        """
        Return the value at the address.

        :param int index: Address to read from
        :param int value: Value to write
        :raises IndexError: If address is out of bounds for block
        :rtype: int
        :return: Value at address (8 bit)
        """
        b = self.getBlock(addr)
        return b.get(addr)

    def readWord(self, addr: int) -> int:
        """
        Return the value at the address.

        :param int index: Address to read from
        :raises IndexError: If address is out of bounds for block
        :rtype: int
        :return: Value at address (16 bit)
        """
        return (self.read(addr + 1) << 8) + self.read(addr)
